@@ -1,6 +1,9 @@
 #ifndef NVIM_EX_DOCMD_H
 #define NVIM_EX_DOCMD_H
 
+#include <stdbool.h>
+
+#include "nvim/buffer_defs.h"
 #include "nvim/ex_cmds_defs.h"
 #include "nvim/globals.h"
 
@@ -12,13 +15,26 @@
 #define DOCMD_EXCRESET  0x10      // reset exception environment (for debugging
 #define DOCMD_KEEPLINE  0x20      // keep typed line for repeating with "."
 
-/* defines for eval_vars() */
+// defines for eval_vars()
 #define VALID_PATH              1
 #define VALID_HEAD              2
 
-/* Values for exmode_active (0 is no exmode) */
-#define EXMODE_NORMAL           1
-#define EXMODE_VIM              2
+// Whether a command index indicates a user command.
+#define IS_USER_CMDIDX(idx) ((int)(idx) < 0)
+
+// Structure used to save the current state.  Used when executing Normal mode
+// commands while in any other mode.
+typedef struct {
+  int save_msg_scroll;
+  int save_restart_edit;
+  bool save_msg_didout;
+  int save_State;
+  bool save_finish_op;
+  long save_opcount;
+  int save_reg_executing;
+  bool save_pending_end_reg_executing;
+  tasave_T tabuf;
+} save_state_T;
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "ex_docmd.h.generated.h"
