@@ -1,9 +1,9 @@
 #ifndef NVIM_MEMLINE_H
 #define NVIM_MEMLINE_H
 
+#include "nvim/buffer_defs.h"
+#include "nvim/pos.h"
 #include "nvim/types.h"
-#include "nvim/pos.h" // for pos_T, linenr_T, colnr_T
-#include "nvim/buffer_defs.h" // for buf_T
 
 #ifdef INCLUDE_GENERATED_DECLARATIONS
 # include "memline.h.generated.h"
