@@ -1,14 +1,17 @@
 local helpers = require('test.functional.helpers')(after_each)
 
+local assert_alive = helpers.assert_alive
 local command = helpers.command
 local feed_command = helpers.feed_command
+local feed = helpers.feed
 local eval = helpers.eval
 local eq = helpers.eq
 local run = helpers.run
 local funcs = helpers.funcs
 local nvim_prog = helpers.nvim_prog
-local redir_exec = helpers.redir_exec
-local wait = helpers.wait
+local pcall_err = helpers.pcall_err
+local exec_capture = helpers.exec_capture
+local poke_eventloop = helpers.poke_eventloop
 
 describe('v:exiting', function()
   local cid
@@ -22,38 +25,43 @@ describe('v:exiting', function()
     eq(1, eval('v:exiting is v:null'))
   end)
 
+  local function test_exiting(setup_fn)
+    local function on_setup()
+      command('autocmd VimLeavePre * call rpcrequest('..cid..', "exit", "VimLeavePre")')
+      command('autocmd VimLeave    * call rpcrequest('..cid..', "exit", "VimLeave")')
+      setup_fn()
+    end
+    local requests_args = {}
+    local function on_request(name, args)
+      eq('exit', name)
+      table.insert(requests_args, args)
+      eq(0, eval('v:exiting'))
+      return ''
+    end
+    run(on_request, nil, on_setup)
+    eq({{'VimLeavePre'}, {'VimLeave'}}, requests_args)
+  end
+
   it('is 0 on normal exit', function()
-    local function on_setup()
-      command('autocmd VimLeavePre * call rpcrequest('..cid..', "")')
-      command('autocmd VimLeave    * call rpcrequest('..cid..', "")')
+    test_exiting(function()
       command('quit')
-    end
-    local function on_request()
-      eq(0, eval('v:exiting'))
-      return ''
-    end
-    run(on_request, nil, on_setup)
+    end)
   end)
-  it('is 0 on exit from ex-mode involving try-catch', function()
-    local function on_setup()
-      command('autocmd VimLeavePre * call rpcrequest('..cid..', "")')
-      command('autocmd VimLeave    * call rpcrequest('..cid..', "")')
-      feed_command('call feedkey("Q")','try', 'call NoFunction()', 'catch', 'echo "bye"', 'endtry', 'quit')
-    end
-    local function on_request()
-      eq(0, eval('v:exiting'))
-      return ''
-    end
-    run(on_request, nil, on_setup)
+
+  it('is 0 on exit from Ex mode involving try-catch vim-patch:8.0.0184', function()
+    test_exiting(function()
+      feed('gQ')
+      feed_command('try', 'call NoFunction()', 'catch', 'echo "bye"', 'endtry', 'quit')
+    end)
   end)
 end)
 
 describe(':cquit', function()
   local function test_cq(cmdline, exit_code, redir_msg)
     if redir_msg then
-      eq('\n' .. redir_msg, redir_exec(cmdline))
-      wait()
-      eq(2, eval("1+1"))  -- Still alive?
+      eq(redir_msg, pcall_err(function() return exec_capture(cmdline) end))
+      poke_eventloop()
+      assert_alive()
     else
       funcs.system({nvim_prog, '-u', 'NONE', '-i', 'NONE', '--headless', '--cmd', cmdline})
       eq(exit_code, eval('v:shell_error'))
@@ -85,14 +93,14 @@ describe(':cquit', function()
   end)
 
   it('exits with redir msg for multiple exit codes after :cquit 1 2', function()
-    test_cq('cquit 1 2', nil, 'E488: Trailing characters: cquit 1 2')
+    test_cq('cquit 1 2', nil, 'nvim_exec2(): Vim(cquit):E488: Trailing characters: 2: cquit 1 2')
   end)
 
   it('exits with redir msg for non-number exit code after :cquit X', function()
-    test_cq('cquit X', nil, 'E488: Trailing characters: cquit X')
+    test_cq('cquit X', nil, 'nvim_exec2(): Vim(cquit):E488: Trailing characters: X: cquit X')
   end)
 
   it('exits with redir msg for negative exit code after :cquit -1', function()
-    test_cq('cquit -1', nil, 'E488: Trailing characters: cquit -1')
+    test_cq('cquit -1', nil, 'nvim_exec2(): Vim(cquit):E488: Trailing characters: -1: cquit -1')
   end)
 end)
