@@ -2,7 +2,6 @@ local helpers = require('test.functional.helpers')(after_each)
 local Screen = require('test.functional.ui.screen')
 local clear = helpers.clear
 local command = helpers.command
-local curbufmeths = helpers.curbufmeths
 local eq = helpers.eq
 local eval = helpers.eval
 local feed_command = helpers.feed_command
@@ -17,7 +16,7 @@ local source = helpers.source
 local poke_eventloop = helpers.poke_eventloop
 local nvim = helpers.nvim
 local sleep = helpers.sleep
-local nvim_dir = helpers.nvim_dir
+local testprg = helpers.testprg
 local assert_alive = helpers.assert_alive
 
 local default_text = [[
@@ -66,7 +65,6 @@ local function common_setup(screen, inccommand, text)
     command("syntax on")
     command("set nohlsearch")
     command("hi Substitute guifg=red guibg=yellow")
-    command("set display-=msgsep")
     screen:attach()
     screen:set_default_attr_ids({
       [1]  = {foreground = Screen.colors.Fuchsia},
@@ -123,13 +121,11 @@ describe(":substitute, inccommand=split interactivity", function()
   it("no preview if invoked by feedkeys()", function()
     -- in a script...
     source([[:call feedkeys(":%s/tw/MO/g\<CR>")]])
-    poke_eventloop()
     -- or interactively...
-    feed([[:call feedkeys(":%s/tw/MO/g\<CR>")<CR>]])
-    poke_eventloop()
+    feed([[:call feedkeys(":%s/bs/BUU/g\<lt>CR>")<CR>]])
     eq(1, eval("bufnr('$')"))
     -- sanity check: assert the buffer state
-    expect(default_text:gsub("tw", "MO"))
+    expect(default_text:gsub("tw", "MO"):gsub("bs", "BUU"))
   end)
 end)
 
@@ -144,11 +140,11 @@ describe(":substitute, 'inccommand' preserves", function()
     feed_command("ls")
 
     screen:expect([[
+      BAC                           |
       {15:~                             }|
       {15:~                             }|
       {15:~                             }|
-      {15:~                             }|
-      {15:~                             }|
+      {11:                              }|
       :ls                           |
         1 %a + "[No Name]"          |
                 line 1              |
@@ -178,11 +174,14 @@ describe(":substitute, 'inccommand' preserves", function()
     it("'undolevels' (inccommand="..case..")", function()
       feed_command("set undolevels=139")
       feed_command("setlocal undolevels=34")
+      feed_command("split")  -- Show the buffer in multiple windows
       feed_command("set inccommand=" .. case)
       insert("as")
-      feed(":%s/as/glork/<enter>")
-      eq(meths.get_option('undolevels'), 139)
-      eq(curbufmeths.get_option('undolevels'), 34)
+      feed(":%s/as/glork/")
+      poke_eventloop()
+      feed("<enter>")
+      eq(meths.get_option_value('undolevels', {scope='global'}), 139)
+      eq(meths.get_option_value('undolevels', {buf=0}), 34)
     end)
   end
 
@@ -257,42 +256,6 @@ describe(":substitute, 'inccommand' preserves", function()
     end)
   end
 
-  for _, case in pairs{"", "split", "nosplit"} do
-    it("visual selection for non-previewable command (inccommand="..case..") #5888", function()
-      local screen = Screen.new(30,10)
-      common_setup(screen, case, default_text)
-      feed('1G2V')
-
-      feed(':s')
-      screen:expect([[
-        {vis:Inc substitution on}           |
-        t{vis:wo lines}                     |
-                                      |
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        :'<,'>s^                       |
-      ]])
-
-      feed('o')
-      screen:expect([[
-        {vis:Inc substitution on}           |
-        t{vis:wo lines}                     |
-                                      |
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        {15:~                             }|
-        :'<,'>so^                      |
-      ]])
-    end)
-  end
-
   for _, case in ipairs({'', 'split', 'nosplit'}) do
     it('previous substitute string ~ (inccommand='..case..') #12109', function()
       local screen = Screen.new(30,10)
@@ -381,7 +344,7 @@ describe(":substitute, 'inccommand' preserves undo", function()
   }
 
   local function test_sub(substring, split, redoable)
-    clear()
+    command('bwipe!')
     feed_command("set inccommand=" .. split)
 
     insert("1")
@@ -407,7 +370,7 @@ describe(":substitute, 'inccommand' preserves undo", function()
   end
 
   local function test_notsub(substring, split, redoable)
-    clear()
+    command('bwipe!')
     feed_command("set inccommand=" .. split)
 
     insert("1")
@@ -441,7 +404,7 @@ describe(":substitute, 'inccommand' preserves undo", function()
 
 
   local function test_threetree(substring, split)
-    clear()
+    command('bwipe!')
     feed_command("set inccommand=" .. split)
 
     insert("1")
@@ -493,6 +456,8 @@ describe(":substitute, 'inccommand' preserves undo", function()
       2]])
   end
 
+  before_each(clear)
+
   it("at a non-leaf of the undo tree", function()
    for _, case in pairs(cases) do
      for _, str in pairs(substrings) do
@@ -1227,6 +1192,8 @@ describe(":substitute, inccommand=split", function()
   end)
 
   it("deactivates if 'redrawtime' is exceeded #5602", function()
+    -- prevent redraws from 'incsearch'
+    meths.set_option('incsearch', false)
     -- Assert that 'inccommand' is ENABLED initially.
     eq("split", eval("&inccommand"))
     -- Set 'redrawtime' to minimal value, to ensure timeout is triggered.
@@ -1343,6 +1310,108 @@ describe(":substitute, inccommand=split", function()
     ]])
   end)
 
+  it([[preview changes correctly with c_CTRL-R_= and c_CTRL-\_e]], function()
+    feed('gg')
+    feed(":1,2s/t/X")
+    screen:expect([[
+      Inc subs{12:X}itution on           |
+      {12:X}wo lines                     |
+      Inc substitution on           |
+      two lines                     |
+                                    |
+      {11:[No Name] [+]                 }|
+      |1| Inc subs{12:X}itution on       |
+      |2| {12:X}wo lines                 |
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {10:[Preview]                     }|
+      :1,2s/t/X^                     |
+    ]])
+
+    feed([[<C-R>='Y']])
+    -- preview should be unchanged during c_CTRL-R_= editing
+    screen:expect([[
+      Inc subs{12:X}itution on           |
+      {12:X}wo lines                     |
+      Inc substitution on           |
+      two lines                     |
+                                    |
+      {11:[No Name] [+]                 }|
+      |1| Inc subs{12:X}itution on       |
+      |2| {12:X}wo lines                 |
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {10:[Preview]                     }|
+      ={1:'Y'}^                          |
+    ]])
+
+    feed('<CR>')
+    -- preview should be changed by the result of the expression
+    screen:expect([[
+      Inc subs{12:XY}itution on          |
+      {12:XY}wo lines                    |
+      Inc substitution on           |
+      two lines                     |
+                                    |
+      {11:[No Name] [+]                 }|
+      |1| Inc subs{12:XY}itution on      |
+      |2| {12:XY}wo lines                |
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {10:[Preview]                     }|
+      :1,2s/t/XY^                    |
+    ]])
+
+    feed([[<C-\>e'echo']])
+    -- preview should be unchanged during c_CTRL-\_e editing
+    screen:expect([[
+      Inc subs{12:XY}itution on          |
+      {12:XY}wo lines                    |
+      Inc substitution on           |
+      two lines                     |
+                                    |
+      {11:[No Name] [+]                 }|
+      |1| Inc subs{12:XY}itution on      |
+      |2| {12:XY}wo lines                |
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {10:[Preview]                     }|
+      ={1:'echo'}^                       |
+    ]])
+
+    feed('<CR>')
+    -- preview should be cleared if command is changed to a non-previewable one
+    screen:expect([[
+      Inc substitution on           |
+      two lines                     |
+      Inc substitution on           |
+      two lines                     |
+                                    |
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      {15:~                             }|
+      :echo^                         |
+    ]])
+  end)
+
 end)
 
 describe("inccommand=nosplit", function()
@@ -1403,14 +1472,14 @@ describe("inccommand=nosplit", function()
     -- non-modifier prefix
     feed(':silent tabedit %s/tw/to')
     screen:expect([[
+      Inc substitution on |
       two lines           |
       Inc substitution on |
       two lines           |
                           |
       {15:~                   }|
       {15:~                   }|
-      {15:~                   }|
-      {15:~                   }|
+      {11:                    }|
       :silent tabedit %s/t|
       w/to^                |
     ]])
@@ -1646,10 +1715,12 @@ end)
 
 describe("'inccommand' and :cnoremap", function()
   local cases = { "",  "split", "nosplit" }
+  local screen
 
-  local function refresh(case)
+  local function refresh(case, visual)
     clear()
-    common_setup(nil, case, default_text)
+    screen = visual and Screen.new(50,10) or nil
+    common_setup(screen, case, default_text)
   end
 
   it('work with remapped characters', function()
@@ -1706,13 +1777,15 @@ describe("'inccommand' and :cnoremap", function()
 
   it('still works with a broken mapping', function()
     for _, case in pairs(cases) do
-      refresh(case)
+      refresh(case, true)
       feed_command("cnoremap <expr> x execute('bwipeout!')[-1].'x'")
 
       feed(":%s/tw/tox<enter>")
+      screen:expect{any=[[{14:^E565:]]}
+      feed('<c-c>')
 
       -- error thrown b/c of the mapping
-      neq(nil, eval('v:errmsg'):find('^E523:'))
+      neq(nil, eval('v:errmsg'):find('^E565:'))
       expect([[
       Inc substitution on
       toxo lines
@@ -1858,26 +1931,26 @@ describe("'inccommand' split windows", function()
     feed_command("split")
     feed(":%s/tw")
     screen:expect([[
-      Inc substitution on {10:│}Inc substitution on|
-      {12:tw}o lines           {10:│}{12:tw}o lines          |
-                          {10:│}                   |
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {11:[No Name] [+]       }{10:│}{15:~                  }|
-      Inc substitution on {10:│}{15:~                  }|
-      {12:tw}o lines           {10:│}{15:~                  }|
-                          {10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
+      Inc substitution on │Inc substitution on|
+      {12:tw}o lines           │{12:tw}o lines          |
+                          │                   |
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {11:[No Name] [+]       }│{15:~                  }|
+      Inc substitution on │{15:~                  }|
+      {12:tw}o lines           │{15:~                  }|
+                          │{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
       {10:[No Name] [+]        [No Name] [+]      }|
       |2| {12:tw}o lines                           |
       {15:~                                       }|
@@ -1897,20 +1970,20 @@ describe("'inccommand' split windows", function()
 
     feed(":%s/tw")
     screen:expect([[
-      Inc substitution on {10:│}Inc substitution on|
-      {12:tw}o lines           {10:│}{12:tw}o lines          |
-                          {10:│}                   |
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
-      {15:~                   }{10:│}{15:~                  }|
+      Inc substitution on │Inc substitution on|
+      {12:tw}o lines           │{12:tw}o lines          |
+                          │                   |
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
+      {15:~                   }│{15:~                  }|
       {11:[No Name] [+]        }{10:[No Name] [+]      }|
       Inc substitution on                     |
       {12:tw}o lines                               |
@@ -1981,6 +2054,16 @@ describe("'inccommand' split windows", function()
     end
   end)
 
+  it("don't open if there's not enough room", function()
+    refresh()
+    screen:try_resize(40, 3)
+    feed("gg:%s/tw")
+    screen:expect([[
+      Inc substitution on                     |
+      {12:tw}o lines                               |
+      :%s/tw^                                  |
+    ]])
+  end)
 end)
 
 describe("'inccommand' with 'gdefault'", function()
@@ -2576,6 +2659,7 @@ describe(":substitute", function()
 
     feed("\\rѫ ab   \\rXXXX")
     screen:expect([[
+      7 8 9                         |
       K L M                         |
       {12:JLKR £}                        |
       {12:ѫ ab   }                       |
@@ -2587,8 +2671,7 @@ describe(":substitute", function()
       {12:ѫ ab   }                       |
       {11:[No Name] [+]                 }|
       | 7| {12:JLKR £}                   |
-      | 8|{12: ѫ ab   }                  |
-      {10:[Preview]                     }|
+      {11:                              }|
       :%s/[a-z]/JLKR £\rѫ ab   \rXXX|
       X^                             |
     ]])
@@ -2795,7 +2878,7 @@ it(':substitute with inccommand during :terminal activity', function()
     clear()
 
     command("set cmdwinheight=3")
-    feed([[:terminal "]]..nvim_dir..[[/shell-test" REP 5000 xxx<cr>]])
+    feed(([[:terminal "%s" REP 5000 xxx<cr>]]):format(testprg('shell-test')))
     command('file term')
     feed('G')  -- Follow :terminal output.
     command('new')
@@ -2805,7 +2888,8 @@ it(':substitute with inccommand during :terminal activity', function()
     feed('gg')
     feed(':%s/foo/ZZZ')
     sleep(20)  -- Allow some terminal activity.
-    helpers.poke_eventloop()
+    poke_eventloop()
+    screen:sleep(0)
     screen:expect_unchanged()
   end)
 end)
@@ -2837,6 +2921,115 @@ it(':substitute with inccommand, timer-induced :redraw #9777', function()
   ]])
 end)
 
+it(':substitute with inccommand, allows :redraw before first separator is typed #18857', function()
+  local screen = Screen.new(30,6)
+  clear()
+  common_setup(screen, 'split', 'foo bar baz\nbar baz fox\nbar foo baz')
+  command('hi! link NormalFloat CursorLine')
+  local float_buf = meths.create_buf(false, true)
+  meths.open_win(float_buf, false, {
+    relative = 'editor', height = 1, width = 5, row = 3, col = 0, focusable = false,
+  })
+  feed(':%s')
+  screen:expect([[
+    foo bar baz                   |
+    bar baz fox                   |
+    bar foo baz                   |
+    {16:     }{15:                         }|
+    {15:~                             }|
+    :%s^                           |
+  ]])
+  meths.buf_set_lines(float_buf, 0, -1, true, {'foo'})
+  command('redraw')
+  screen:expect([[
+    foo bar baz                   |
+    bar baz fox                   |
+    bar foo baz                   |
+    {16:foo  }{15:                         }|
+    {15:~                             }|
+    :%s^                           |
+  ]])
+end)
+
+it(':substitute with inccommand, does not crash if range contains invalid marks', function()
+  local screen = Screen.new(30, 6)
+  clear()
+  common_setup(screen, 'split', 'test')
+  feed([[:'a,'bs]])
+  screen:expect([[
+    test                          |
+    {15:~                             }|
+    {15:~                             }|
+    {15:~                             }|
+    {15:~                             }|
+    :'a,'bs^                       |
+  ]])
+  -- v:errmsg shouldn't be set either before the first separator is typed
+  eq('', eval('v:errmsg'))
+  feed('/')
+  screen:expect([[
+    test                          |
+    {15:~                             }|
+    {15:~                             }|
+    {15:~                             }|
+    {15:~                             }|
+    :'a,'bs/^                      |
+  ]])
+end)
+
+it(':substitute with inccommand, no unnecessary redraw if preview is not shown', function()
+  local screen = Screen.new(60, 6)
+  clear()
+  common_setup(screen, 'split', 'test')
+  feed(':ls<CR>')
+  screen:expect([[
+    test                                                        |
+    {15:~                                                           }|
+    {11:                                                            }|
+    :ls                                                         |
+      1 %a + "[No Name]"                    line 1              |
+    {13:Press ENTER or type command to continue}^                     |
+  ]])
+  feed(':s')
+  -- no unnecessary redraw, so messages are still shown
+  screen:expect([[
+    test                                                        |
+    {15:~                                                           }|
+    {11:                                                            }|
+    :ls                                                         |
+      1 %a + "[No Name]"                    line 1              |
+    :s^                                                          |
+  ]])
+  feed('o')
+  screen:expect([[
+    test                                                        |
+    {15:~                                                           }|
+    {11:                                                            }|
+    :ls                                                         |
+      1 %a + "[No Name]"                    line 1              |
+    :so^                                                         |
+  ]])
+  feed('<BS>')
+  screen:expect([[
+    test                                                        |
+    {15:~                                                           }|
+    {11:                                                            }|
+    :ls                                                         |
+      1 %a + "[No Name]"                    line 1              |
+    :s^                                                          |
+  ]])
+  feed('/test')
+  -- now inccommand is shown, so screen is redrawn
+  screen:expect([[
+    {12:test}                                                        |
+    {15:~                                                           }|
+    {15:~                                                           }|
+    {15:~                                                           }|
+    {15:~                                                           }|
+    :s/test^                                                     |
+  ]])
+end)
+
 it(":substitute doesn't crash with inccommand, if undo is empty #12932", function()
   local screen = Screen.new(10,5)
   clear()
@@ -2857,6 +3050,43 @@ it(":substitute doesn't crash with inccommand, if undo is empty #12932", functio
   assert_alive()
 end)
 
+it(':substitute with inccommand works properly if undo is not synced #20029', function()
+  local screen = Screen.new(30, 6)
+  clear()
+  common_setup(screen, 'nosplit', 'foo\nbar\nbaz')
+  meths.set_keymap('x', '<F2>', '<Esc>`<Oaaaaa asdf<Esc>`>obbbbb asdf<Esc>V`<k:s/asdf/', {})
+  feed('gg0<C-V>lljj<F2>')
+  screen:expect([[
+    aaaaa                         |
+    foo                           |
+    bar                           |
+    baz                           |
+    bbbbb                         |
+    :'<,'>s/asdf/^                 |
+  ]])
+  feed('hjkl')
+  screen:expect([[
+    aaaaa {12:hjkl}                    |
+    foo                           |
+    bar                           |
+    baz                           |
+    bbbbb {12:hjkl}                    |
+    :'<,'>s/asdf/hjkl^             |
+  ]])
+  feed('<CR>')
+  expect([[
+    aaaaa hjkl
+    foo
+    bar
+    baz
+    bbbbb hjkl]])
+  feed('u')
+  expect([[
+    foo
+    bar
+    baz]])
+end)
+
 it('long :%s/ with inccommand does not collapse cmdline', function()
   local screen = Screen.new(10,5)
   clear()
@@ -2865,10 +3095,23 @@ it('long :%s/ with inccommand does not collapse cmdline', function()
   feed(':%s/AAAAAAA', 'A', 'A', 'A', 'A', 'A', 'A', 'A', 'A', 'A', 'A', 'A',
     'A', 'A', 'A', 'A', 'A', 'A', 'A', 'A', 'A')
   screen:expect([[
-    {15:~           }|
-    {15:~           }|
+                |
+    {11:            }|
     :%s/AAAAAAAA|
     AAAAAAAAAAAA|
     AAAAAAA^     |
   ]])
 end)
+
+it("with 'inccommand' typing :filter doesn't segfault or leak memory #19057", function()
+  clear()
+  common_setup(nil, 'nosplit')
+  feed(':filter s')
+  assert_alive()
+  feed(' ')
+  assert_alive()
+  feed('h')
+  assert_alive()
+  feed('i')
+  assert_alive()
+end)
